mod error;

use super::Frequency as RequestedFrequencyType;
use rubato::{FftFixedInOut, Resampler};
use symphonia::core::{
    audio::{Layout, SampleBuffer},
    codecs::{Decoder, DecoderOptions, CODEC_TYPE_NULL},
    errors::Error as SymError,
    formats::{FormatOptions, FormatReader},
    io::MediaSourceStream,
    meta::MetadataOptions,
    probe::Hint, sample
};
use tool_helper::{Error, Input};

#[derive(Copy, Clone, PartialEq)]
pub enum Orality {
    Stereo,
    Mono,
}

impl Orality {
    fn as_channels(&self) -> usize {
        match self {
            Orality::Mono   => 1,
            Orality::Stereo => 2,
        }
    }
}

pub struct PreparedAudioSamples {
    samples:   Vec::<i16>,
    orality:   Orality,
    frequency: u32,
}

impl PreparedAudioSamples {
    pub fn new(samples: Vec<i16>, channels: usize, frequency: u32) -> Result<PreparedAudioSamples, Error> {
        let orality = match channels {
            0 => return Err(Error::from_str("Input file has no audio channels")),
            1 => Orality::Mono,
            2 => Orality::Stereo,
            _ => return Err(Error::from_str("Only Mono and Stereo input are supported")),
        };
        Ok(PreparedAudioSamples{samples, orality, frequency})
    }

    pub fn samples(&self) -> &Vec::<i16> {
        &self.samples
    }

    pub fn orality(&self) -> Orality {
        self.orality.clone()
    }
}

struct InternalAudioSamples {
    planar_samples: Vec<Vec<f32>>,
    frequency:      u32,
}

impl InternalAudioSamples {
    pub fn new(planar_samples: Vec<Vec<f32>>, frequency: u32) -> Result<InternalAudioSamples, Error> {
        if planar_samples.len() < 1 {
            // TODO: Make better text
            Err(Error::from_str("Audio samples can not be empty"))
        }

        else {
            Ok(InternalAudioSamples{planar_samples, frequency})
        }
    }

    pub fn sample_len(&self) -> usize {
        self.planar_samples[0].len()
    }

    pub fn channels(&self) -> usize {
        self.planar_samples.len()
    }

    pub fn planar_slices(&self) -> Vec<&[f32]> {
        let mut planar_slices = Vec::new();

        for channel in &self.planar_samples {
            planar_slices.push(channel.as_slice());
        }
        planar_slices
    }
}

pub fn load_as_i16_audio(input: Input, freq_type: RequestedFrequencyType) -> Result<PreparedAudioSamples, Error> {
    let raw_audio = load_raw_audio(input)?;
    let raw_audio = resample(raw_audio, 37_800)?;

    test_write_wav(down_sample_interleave(raw_audio)?)?;
    Err(Error::not_implemented("Resampling not implemented"))
}

fn test_write_wav(audio_samples: PreparedAudioSamples) -> Result<(), Error> {
    let spec = hound::WavSpec {
        channels: 2,
        sample_rate: 37_800,
        bits_per_sample: 16,
        sample_format: hound::SampleFormat::Int,
    };

    let mut file = hound::WavWriter::create("planschi.wav", spec).unwrap();
    for sample in audio_samples.samples {
        file.write_sample(sample)?;
    }
    file.finalize()?;
    Ok(())
}

fn load_raw_audio(input: Input) -> Result<InternalAudioSamples, Error> {
    let media_stream = MediaSourceStream::new(Box::new(load_to_ram(input)?), Default::default());
    let format       = symphonia::default::get_probe().format(&Hint::new(), media_stream, &FormatOptions::default(), &MetadataOptions::default()).map_err(error::probe)?.format;
    let track        = format.tracks().iter().find(|t| t.codec_params.codec != CODEC_TYPE_NULL).ok_or_else(error::find_track)?;
    
    // Create a decoder for the track.
    let decoder  = symphonia::default::get_codecs().make(&track.codec_params, &DecoderOptions::default()).map_err(error::decoder)?;
    let track_id = track.id;

    decode(format, decoder, track_id)
}

fn decode(mut format: Box<dyn FormatReader>, mut decoder: Box<dyn Decoder>, track_id: u32) -> Result<InternalAudioSamples, Error> {
    let mut samples       = Vec::new();
    let mut channel_count = 0;
    let mut frequency     = 0;
    let mut read_buffer   = None;

    loop {
        // Get the next packet from the media format.
        let packet = match format.next_packet() {
            Ok(packet) => packet,
            Err(err) => {
                if let SymError::IoError(io_err) = &err {
                    if io_err.kind() == std::io::ErrorKind::UnexpectedEof {
                        return InternalAudioSamples::new(samples, frequency);
                    }                    
                }
                return Err(error::next_packet(err));
            }
        };

        // Consume any new metadata that has been read since the last packet.
        format.metadata().skip_to_latest();

        // If the packet does not belong to the selected track, skip over it.
        if packet.track_id() != track_id {
            continue;
        }

        // Decode the packet into audio samples.
        let packet = decoder.decode(&packet).map_err(error::decode)?;
        if read_buffer.is_none() {
            let duration = packet.capacity() as u64;
            let specs    = packet.spec();

            channel_count = specs.channels.count();
            frequency     = specs.rate;
            read_buffer   = Some(SampleBuffer::<f32>::new(duration, packet.spec().clone()));
            
            for _ in 0..channel_count {
                samples.push(Vec::new());
            }
        }

        if let Some(read_buffer) = &mut read_buffer {
            read_buffer.copy_planar_ref(packet);
            let     cur_samples = read_buffer.samples();
            let mut cur_samples = cur_samples.chunks(cur_samples.len()/channel_count);

            for dst_sample in &mut samples {
                dst_sample.extend(cur_samples.next().ok_or_else(|| Error::from_str("Not enough channels in input as expected"))?);
            }
        }        
    }
}

fn resample(input: InternalAudioSamples, target_frequency: u32) -> Result<InternalAudioSamples, Error> {
    const HIGH_QUALITY_CHUNKS:usize = (1024*10)*100;
    fn process_partial(input_option: Option<&[&[f32]]>, resampler: &mut FftFixedInOut<f32>, planar_output: &mut Vec<Vec<f32>>) -> Result<(), Error> {
        let new_samples = resampler.process_partial(input_option, None).map_err(error::resample)?;
        for (channel, channel_samples) in new_samples.into_iter().enumerate() {
            planar_output[channel].extend(channel_samples.iter());
        }
        Ok(())
    }

    let chunk_size         = HIGH_QUALITY_CHUNKS;
    let mut planar_input   = input.planar_slices();
    let mut resampler      = FftFixedInOut::<f32>::new(input.frequency as usize, target_frequency as usize, chunk_size, input.channels()).map_err(error::resampler_construction)?;
    let     delay          = resampler.output_delay();
    let mut sample_len     = input.sample_len();
    let     new_sample_len = (sample_len as f64*(target_frequency as f64/input.frequency as f64)) as usize;
    let mut planar_output  = {
        let mut planar_output = Vec::new();

        for _ in 0..planar_input.len() {
            planar_output.push(Vec::<f32>::new());
        }
        planar_output
    };

    loop {
        let next_input_frames = resampler.input_frames_next();
        if next_input_frames > sample_len {
            if sample_len > 0 {
                // Still frames left
                process_partial(Some(&planar_input), &mut resampler, &mut planar_output)?;
            }
            break;
        }

        let new_samples = resampler.process(&planar_input, None).map_err(error::resample)?;
        for (channel, slice) in planar_input.iter_mut().enumerate() {
            *slice = &slice[next_input_frames..];
            planar_output[channel].extend(new_samples[channel].iter());
        }
        sample_len -= next_input_frames;
    }

    if planar_output[0].len() < delay + new_sample_len {
        // Flush
        process_partial(None, &mut resampler, &mut planar_output)?;
    }

    for channel in &mut planar_output {
        let start = delay;
        let end   = start + new_sample_len;
        *channel = channel[start..end].into();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    let spec = hound::WavSpec {
        channels: 1,
        sample_rate: 37_800,
        bits_per_sample: 32,
        sample_format: hound::SampleFormat::Float,
    };

    let mut file = hound::WavWriter::create("planschi.wav", spec).unwrap();
    for sample in &planar_output[0] {
        file.write_sample(*sample)?;
    }
    file.finalize()?;

    InternalAudioSamples::new(planar_output, target_frequency)
}

fn down_sample_interleave(input: InternalAudioSamples) -> Result<PreparedAudioSamples, Error> {
    Err(Error::not_implemented("down_sample_interleave"))
}

fn load_to_ram(mut input: Input) -> Result<std::io::Cursor<Vec<u8>>, Error> {
    let mut buffer = Vec::default();
    
    input.read_to_end(&mut buffer)?;
    Ok(std::io::Cursor::new(buffer))
}